from typing import Dict
from typing import List
from typing import Union

import numpy as np

from sonusai.mixture.mixdb import Augmentation
from sonusai.mixture.mixdb import Augmentations
from sonusai.mixture.mixdb import evaluate_random_rule


def get_augmentations(rules: Union[List[Dict], Dict]) -> Augmentations:
    """Generate augmentations from list of input rules."""

    from sonusai.utils import dataclass_from_dict

    processed_rules = list()
    if not isinstance(rules, list):
        rules = [rules]

    for rule in rules:
        expand_rules(processed_rules, rule)

    processed_rules = randomize_rules(processed_rules)

    augmentations = list()
    for processed_rule in processed_rules:
        augmentations.append(dataclass_from_dict(Augmentation, processed_rule))

    return augmentations


def expand_rules(out_rules: List[Dict], in_rule: dict) -> None:
    """Expand rules."""
    from copy import deepcopy
    from numbers import Number

    from sonusai import SonusAIError
    from sonusai.mixture import VALID_AUGMENTATIONS

    for key, value in list(in_rule.items()):
        if value is None:
            del in_rule[key]

    # replace old 'eq' rule with new 'eq1' rule to allow both for backward compatibility
    in_rule = {'eq1' if key == 'eq' else key: value for key, value in in_rule.items()}

    for key in in_rule:
        if key not in VALID_AUGMENTATIONS:
            nice_list = '\n'.join([f'  {item}' for item in VALID_AUGMENTATIONS])
            raise SonusAIError(f'Invalid augmentation: {key}.\nValid augmentations are:\n{nice_list}')

        if key in ['eq1', 'eq2', 'eq3']:
            # eq must be a list of length 3 or a list of length 3 lists
            valid = True
            multiple = False
            if isinstance(in_rule[key], list):
                if any(isinstance(el, list) for el in in_rule[key]):
                    multiple = True
                    for value in in_rule[key]:
                        if not isinstance(value, list) or len(value) != 3:
                            valid = False
                else:
                    if len(in_rule[key]) != 3:
                        valid = False
            else:
                valid = False

            if not valid:
                raise SonusAIError(f'Invalid augmentation value for {key}: {in_rule[key]}')

            if multiple:
                for value in in_rule[key]:
                    expanded_rule = deepcopy(in_rule)
                    expanded_rule[key] = deepcopy(value)
                    expand_rules(out_rules, expanded_rule)
                return

        elif key in ['count', 'mixup']:
            pass

        else:
            if isinstance(in_rule[key], list):
                for value in in_rule[key]:
                    if isinstance(value, list):
                        raise SonusAIError(f'Invalid augmentation value for {key}: {in_rule[key]}')
                    expanded_rule = deepcopy(in_rule)
                    expanded_rule[key] = deepcopy(value)
                    expand_rules(out_rules, expanded_rule)
                return
            elif not isinstance(in_rule[key], Number):
                if not in_rule[key].startswith('rand'):
                    raise SonusAIError(f'Invalid augmentation value for {key}: {in_rule[key]}')

    out_rules.append(in_rule)


def randomize_rules(in_rules: List[Dict]) -> List[Dict]:
    """Randomize rules."""
    out_rules = list()
    for in_rule in in_rules:
        if rule_has_rand(in_rule):
            count = 1
            if 'count' in in_rule and in_rule['count'] is not None:
                count = in_rule['count']
                del in_rule['count']
            for i in range(count):
                out_rules.append(generate_random_rule(in_rule))
        else:
            out_rules.append(in_rule)
    return out_rules


def generate_random_rule(in_rule: dict) -> dict:
    """Generate a new rule from a rule that contains 'rand' directives."""
    from copy import deepcopy

    out_rule = deepcopy(in_rule)
    for key in out_rule:
        out_rule[key] = evaluate_random_rule(str(out_rule[key]))

        # convert eq values from strings to numbers
        if key in ['eq1', 'eq2', 'eq3']:
            for n in range(3):
                if isinstance(out_rule[key][n], str):
                    out_rule[key][n] = eval(out_rule[key][n])

    return out_rule


def rule_has_rand(rule: dict) -> bool:
    """Determine if any keys in the given rule contain 'rand'"""
    for key in rule:
        if 'rand' in str(rule[key]):
            return True

    return False


def apply_augmentation(audio_in: np.ndarray,
                       augmentation: Augmentation,
                       length_common_denominator: int = 1) -> np.ndarray:
    """Use sox to apply augmentations to audio data."""

    import sox

    from sonusai import SonusAIError
    from sonusai.mixture import BIT_DEPTH
    from sonusai.mixture import CHANNEL_COUNT
    from sonusai.mixture import SAMPLE_RATE

    try:
        # Apply augmentations
        tfm = sox.Transformer()
        tfm.set_input_format(rate=SAMPLE_RATE, bits=BIT_DEPTH, channels=CHANNEL_COUNT)
        tfm.set_output_format(rate=SAMPLE_RATE, bits=BIT_DEPTH, channels=CHANNEL_COUNT)

        # TODO
        #  Always normalize and remove normalize from list of available augmentations
        #  Normalize to globally set level (should this be a global config parameter,
        #  or hard-coded into the script?)
        if augmentation.normalize is not None:
            tfm.norm(db_level=augmentation.normalize)

        if augmentation.gain is not None:
            tfm.gain(gain_db=augmentation.gain, normalize=False)

        if augmentation.pitch is not None:
            tfm.pitch(n_semitones=augmentation.pitch / 100)

        if augmentation.tempo is not None:
            factor = augmentation.tempo
            if abs(factor - 1.0) <= 0.1:
                tfm.stretch(factor=factor)
            else:
                tfm.tempo(factor=factor, audio_type='s')

        if augmentation.eq1 is not None:
            tfm.equalizer(frequency=augmentation.eq1[0], width_q=augmentation.eq1[1],
                          gain_db=augmentation.eq1[2])

        if augmentation.eq2 is not None:
            tfm.equalizer(frequency=augmentation.eq2[0], width_q=augmentation.eq2[1],
                          gain_db=augmentation.eq2[2])

        if augmentation.eq3 is not None:
            tfm.equalizer(frequency=augmentation.eq3[0], width_q=augmentation.eq3[1],
                          gain_db=augmentation.eq3[2])

        if augmentation.lpf is not None:
            tfm.lowpass(frequency=augmentation.lpf)

        # Create output data
        audio_out = tfm.build_array(input_array=audio_in, sample_rate_in=SAMPLE_RATE)

        # make sure length is multiple of length_common_denominator
        audio_out = np.pad(array=audio_out,
                           pad_width=(0, get_pad_length(len(audio_out), length_common_denominator)),
                           mode='constant',
                           constant_values=0)

        return audio_out
    except Exception as e:
        raise SonusAIError(f'Error applying {augmentation}: {e}')


def estimate_audio_length(audio_in: np.ndarray,
                          augmentation: Augmentation,
                          length_common_denominator: int = 1) -> int:
    """Estimate the length of audio after augmentation."""
    length = len(audio_in)

    if augmentation.tempo is not None:
        length = int(length // augmentation.tempo)

    length += get_pad_length(length, length_common_denominator)

    return length


def get_mixups(augmentations: Augmentations) -> List[int]:
    return sorted(list(set([augmentation.mixup for augmentation in augmentations])))


def get_augmentation_indices_for_mixup(augmentations: list, mixup: int) -> list:
    indices = list()
    for idx, augmentation in enumerate(augmentations):
        if mixup == augmentation.mixup:
            indices.append(idx)

    return indices


def get_pad_length(length: int, length_common_denominator: int) -> int:
    mod = int(length % length_common_denominator)
    if mod:
        return length_common_denominator - mod
    return 0



"""sonusai mkwav

usage: mkwav [-htn] [-i MIXID] MIXDB

options:
   -h, --help
   -i MIXID, --mixid MIXID      Mixture ID to generate. [default: 0].
   -t, --target                 Write target file.
   -n, --noise                  Write noise file.

The mkwav command creates WAV files from a SonusAI database.

Inputs:
    MIXDB       A SonusAI mixture database JSON file or an HDF5 file containing a SonusAI mixture database.
                The HDF5 file contains:
                    attribute:  mixdb

"""

import numpy as np

from sonusai import logger
from sonusai.mixture import MixtureDatabase


def mkwav(mixdb: MixtureDatabase, mixid: int) -> (np.ndarray, np.ndarray):
    from sonusai.genmix import genmix

    mixture, _, target, noise, _, _ = genmix(mixdb=mixdb, mixid=[mixid])
    return mixture, target, noise


def _write_wav(name: str, data: np.ndarray) -> None:
    import wave

    from sonusai.mixture import CHANNEL_COUNT
    from sonusai.mixture import SAMPLE_BYTES
    from sonusai.mixture import SAMPLE_RATE

    with wave.open(name, mode='w') as f:
        f.setnchannels(CHANNEL_COUNT)
        f.setsampwidth(SAMPLE_BYTES)
        f.setframerate(SAMPLE_RATE)
        f.writeframesraw(data)
        logger.info(f'Wrote {name}')


def main():
    from os.path import basename
    from os.path import splitext

    from docopt import docopt

    import sonusai
    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.mixture import load_mixdb
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    mixdb_name = args['MIXDB']
    mixid = int(args['--mixid'])

    log_name = 'mkwav.log'
    create_file_handler(log_name)
    update_console_handler(False)
    initial_log_messages('mkwav')

    mixdb = load_mixdb(mixdb_name)

    mixture, target, noise = mkwav(mixdb=mixdb, mixid=mixid)

    base_name = basename(splitext(mixdb_name)[0])
    _write_wav(name=f'{base_name}_mixture_{mixid}.wav', data=mixture)
    if args['--target']:
        _write_wav(name=f'{base_name}_target_{mixid}.wav', data=target)
    if args['--noise']:
        _write_wav(name=f'{base_name}_noise_{mixid}.wav', data=noise)


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



# SonusAI general utilities

from sonusai.utils.calculate_input_shape import calculate_input_shape
from sonusai.utils.dataclass_from_dict import dataclass_from_dict
from sonusai.utils.dbfs import dbfs_to_sample
from sonusai.utils.dbfs import sample_to_dbfs
from sonusai.utils.engineering_number import EngineeringNumber
from sonusai.utils.get_label_names import get_label_names
from sonusai.utils.grouper import grouper
from sonusai.utils.human_readable_size import human_readable_size
from sonusai.utils.import_keras_model import import_keras_model
from sonusai.utils.int16_to_float import int16_to_float
from sonusai.utils.onnx_utils import add_sonusai_metadata
from sonusai.utils.onnx_utils import create_onnx_from_keras
from sonusai.utils.onnx_utils import replace_stateful_grus
from sonusai.utils.parallel import initialize_map_func
from sonusai.utils.parallel import p_map
from sonusai.utils.parallel import p_umap
from sonusai.utils.parallel_tqdm import p_tqdm_map
from sonusai.utils.parallel_tqdm import p_tqdm_umap
from sonusai.utils.print_mixture_details import print_class_count
from sonusai.utils.print_mixture_details import print_mixture_details
from sonusai.utils.ranges import consolidate_range
from sonusai.utils.ranges import expand_range
from sonusai.utils.read_mixture_data import read_mixture_data
from sonusai.utils.read_predict_data import read_predict_data
from sonusai.utils.reshape import reshape_inputs
from sonusai.utils.reshape import reshape_truth_predict
from sonusai.utils.seconds_to_hms import seconds_to_hms
from sonusai.utils.stratified_shuffle_split import stratified_shuffle_split_mixid
from sonusai.utils.to_numpy_array import to_numpy_array
from sonusai.utils.trim_docstring import trim_docstring
from sonusai.utils.yes_or_no import yes_or_no



from dataclasses import dataclass
from typing import List
from typing import Union

import numpy as np

from sonusai.mixture.mixdb import Mixture
from sonusai.mixture.mixdb import MixtureDatabase

# NOTE: multiprocessing dictionary is required for run-time performance; using 'partial' is much slower.
MP_DICT = dict()


def build_noise_audios(mixdb: MixtureDatabase, show_progress: bool = False) -> List[List[np.ndarray]]:
    """Build a database of noise audio data."""
    from tqdm import tqdm

    from sonusai.mixture import apply_augmentation

    noise_audios = list()
    for file_index in tqdm(range(len(mixdb.noises)), desc='Read noise audio', disable=not show_progress):
        audio_in = read_audio(name=mixdb.noises[file_index].name)
        noise_audios.append(list())
        for augmentation_index, augmentation in enumerate(mixdb.noise_augmentations):
            noise_audios[-1].append(apply_augmentation(audio_in=audio_in, augmentation=augmentation))

    return noise_audios


def build_target_audios(mixdb: MixtureDatabase, show_progress: bool = False) -> List[np.ndarray]:
    """Build a list of target audio data."""
    from tqdm import tqdm

    from sonusai.mixture import MixtureDatabase
    from sonusai.utils import p_tqdm_map

    MP_DICT['mixdb']: MixtureDatabase = mixdb

    indices = list(range(len(mixdb.targets)))
    progress = tqdm(total=len(indices), desc='Read target audio', disable=not show_progress)
    target_audios = p_tqdm_map(_read_target_audio, indices, progress=progress)
    return target_audios


def _read_target_audio(file_index: int) -> np.ndarray:
    """Parallel target audio reader kernel."""
    mixdb: MixtureDatabase = MP_DICT['mixdb']
    return read_audio(name=mixdb.targets[file_index].name)


def check_audio_files_exist(mixdb: MixtureDatabase) -> None:
    """Walk through all the noise and target audio files in a mixture database ensuring that they exist."""
    from os.path import exists

    from sonusai import SonusAIError
    from sonusai.mixture import tokenized_expandvars

    for file_index in range(len(mixdb.noises)):
        file_name, _ = tokenized_expandvars(mixdb.noises[file_index].name)
        if not exists(file_name):
            raise SonusAIError(f'Could not find {file_name}')

    for file_index in range(len(mixdb.targets)):
        file_name, _ = tokenized_expandvars(mixdb.targets[file_index].name)
        if not exists(file_name):
            raise SonusAIError(f'Could not find {file_name}')


def read_raw_target_audio(mixdb: MixtureDatabase,
                          show_progress: bool = False) -> List[np.ndarray]:
    """Read in all audio data beforehand to avoid reading it multiple times in a loop."""

    from tqdm import tqdm

    from sonusai.utils import p_tqdm_map

    names = [target.name for target in mixdb.targets]
    progress = tqdm(total=len(names), desc='Read target audio', disable=not show_progress)
    raw_target_audio = p_tqdm_map(read_audio, names, progress=progress)
    progress.close()

    return raw_target_audio


def get_target_noise_audio(mixdb: MixtureDatabase,
                           mixture: Mixture,
                           raw_target_audios: List[np.ndarray],
                           raw_noise_audios: List[List[np.ndarray]]) -> (List[np.ndarray], np.ndarray):
    """Apply augmentations and return augmented target and noise data."""

    from sonusai import SonusAIError
    from sonusai.mixture import apply_augmentation

    target_file_index = mixture.target_file_index
    target_augmentation_index = mixture.target_augmentation_index
    if len(target_file_index) != len(target_augmentation_index):
        raise SonusAIError('target_file_index and target_augmentation_index are not the same length')

    if mixture.samples % mixdb.frame_size != 0:
        raise SonusAIError(f'Number of samples in mixture is not a multiple of {mixdb.frame_size}')

    target_audios = list()
    for idx in range(len(mixture.target_file_index)):
        target_augmentation = mixdb.target_augmentations[target_augmentation_index[idx]]

        target_audio = apply_augmentation(audio_in=raw_target_audios[target_file_index[idx]],
                                          augmentation=target_augmentation,
                                          length_common_denominator=mixdb.feature_step_samples)

        target_audio = np.int16(np.round(np.single(target_audio) * mixture.target_snr_gain))
        target_audio = pad_to_samples(audio_in=target_audio, samples=mixture.samples)
        target_audios.append(target_audio)

    raw_noise_audio = raw_noise_audios[mixture.noise_file_index][mixture.noise_augmentation_index]
    noise_audio, _ = get_next_noise(offset_in=mixture.noise_offset,
                                    length=mixture.samples,
                                    audio_in=raw_noise_audio)

    noise_audio = np.int16(np.round(np.single(noise_audio) * mixture.noise_snr_gain))

    return target_audios, noise_audio


def pad_to_samples(audio_in: np.ndarray, samples: int) -> np.ndarray:
    return np.pad(audio_in, (0, samples - len(audio_in)), mode='constant', constant_values=0)


def get_audio_and_truth_t(mixdb: MixtureDatabase,
                          mixture: Mixture,
                          raw_target_audios: List[np.ndarray],
                          raw_noise_audios: List[List[np.ndarray]],
                          compute_truth: bool = True,
                          compute_segsnr: bool = False,
                          frame_based_segsnr: bool = False) \
        -> (np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray):
    from sonusai.mixture import generate_segsnr
    from sonusai.mixture import generate_truth

    target_audios, noise_audio = get_target_noise_audio(mixdb=mixdb,
                                                        mixture=mixture,
                                                        raw_target_audios=raw_target_audios,
                                                        raw_noise_audios=raw_noise_audios)

    truth_t = generate_truth(mixdb=mixdb,
                             mixture=mixture,
                             target_audios=target_audios,
                             noise_audio=noise_audio,
                             compute=compute_truth)

    target_audio = sum(target_audios)

    segsnr = generate_segsnr(mixdb=mixdb,
                             mixture=mixture,
                             target_audio=target_audio,
                             noise_audio=noise_audio,
                             compute=compute_segsnr,
                             frame_based=frame_based_segsnr)

    mixture_audio = target_audio + noise_audio
    return mixture_audio, truth_t, target_audio, noise_audio, segsnr


def get_next_noise(offset_in: int, length: int, audio_in: np.ndarray) -> (np.ndarray, int):
    audio_out = np.take(audio_in, range(offset_in, offset_in + length), mode='wrap')
    offset_out = (offset_in + length) % len(audio_in)
    return audio_out, offset_out


def read_audio(name: str) -> np.ndarray:
    import sox

    from sonusai import SonusAIError
    from sonusai import logger
    from sonusai.mixture import tokenized_expandvars
    from sonusai.mixture import BIT_DEPTH
    from sonusai.mixture import CHANNEL_COUNT
    from sonusai.mixture import SAMPLE_RATE

    expanded_name, _ = tokenized_expandvars(name)

    try:
        # Read in and convert to desired format
        inp = sox.Transformer()
        inp.set_output_format(rate=SAMPLE_RATE, bits=BIT_DEPTH, channels=CHANNEL_COUNT)
        return inp.build_array(input_filepath=expanded_name,
                               sample_rate_in=int(sox.file_info.sample_rate(expanded_name)))

    except Exception as e:
        if name != expanded_name:
            logger.error(f'Error reading {name} (expanded: {expanded_name}): {e}')
        else:
            raise SonusAIError(f'Error reading {name}: {e}')


@dataclass(frozen=True)
class MixtureData:
    mixture: np.ndarray
    target: List[Union[np.ndarray, None]]
    noise: np.ndarray
    feature: np.ndarray
    truth_f: np.ndarray


def get_mixture_data(mixdb: MixtureDatabase, mixid: int) -> MixtureData:
    """Get mixture data assuming nothing has been loaded into memory already."""

    from sonusai.mixture import apply_augmentation
    from sonusai.mixture import get_feature_and_truth_f
    from sonusai.mixture import get_truth_indices_for_target
    from sonusai.mixture import generate_truth

    mixture = mixdb.mixtures[mixid]

    target_audios = list()
    target_truth_indices = list()
    for idx in range(len(mixture.target_file_index)):
        target_name = mixdb.targets[mixture.target_file_index[idx]].name
        target_augmentation = mixdb.target_augmentations[mixture.target_augmentation_index[idx]]

        target_audio = apply_augmentation(audio_in=read_audio(target_name),
                                          augmentation=target_augmentation,
                                          length_common_denominator=mixdb.feature_step_samples)

        target_audio = np.int16(np.round(np.single(target_audio) * mixture.target_snr_gain))
        target_audio = pad_to_samples(audio_in=target_audio, samples=mixture.samples)
        target_audios.append(target_audio)
        target_truth_indices.append(get_truth_indices_for_target(mixdb.targets[mixture.target_file_index[idx]]))

    raw_noise_audios = build_noise_audios(mixdb=mixdb, show_progress=False)
    raw_noise_audio = raw_noise_audios[mixture.noise_file_index][mixture.noise_augmentation_index]
    noise_audio, _ = get_next_noise(offset_in=mixture.noise_offset,
                                    length=mixture.samples,
                                    audio_in=raw_noise_audio)

    noise_audio = np.int16(np.round(np.single(noise_audio) * mixture.noise_snr_gain))

    truth_t = generate_truth(mixdb=mixdb,
                             mixture=mixture,
                             target_audios=target_audios,
                             noise_audio=noise_audio,
                             compute=True)

    mixture_audio = sum(target_audios) + noise_audio

    # Transform target_audio into a list num_classes long such that each entry is the target data per class
    class_audio = list()
    for n in range(mixdb.num_classes):
        class_audio.append(None)
        for idx in range(len(target_audios)):
            if n + 1 in target_truth_indices[idx]:
                if class_audio[n] is None:
                    class_audio[n] = target_audios[idx]
                else:
                    class_audio[n] += target_audios[idx]

    feature, truth_f = get_feature_and_truth_f(mixdb=mixdb,
                                               mixid=mixid,
                                               audio=mixture_audio,
                                               truth_t=truth_t)

    return MixtureData(mixture=mixture_audio,
                       target=class_audio,
                       noise=noise_audio,
                       feature=feature,
                       truth_f=truth_f)



import numpy as np


def reshape_inputs(feature: np.ndarray,
                   truth: np.ndarray,
                   batch_size: int,
                   timesteps: int = 0,
                   flatten: bool = False,
                   add1ch: bool = False) -> (np.ndarray, np.ndarray, tuple, int, np.ndarray, str):
    """Check SonusAI feature and truth data and reshape feature of size frames x strides x bands into
    one of several options:

    If timesteps > 0: (i.e. for recurrent NNs):
      no-flatten, no-channel:   sequences x timesteps x strides       x bands     (4-dim)
      flatten, no-channel:      sequences x timesteps x strides*bands             (3-dim)
      no-flatten, add-1channel: sequences x timesteps x strides       x bands x 1 (5-dim)
      flatten, add-1channel:    sequences x timesteps x strides*bands         x 1 (4-dim)

    If timesteps == 0, then do not add timesteps dimension

    The number of samples is trimmed to be a multiple of batch_size (Keras requirement) for
    both feature and truth.
    Channel is added to last/outer dimension for channel_last support in Keras/TF

    Returns:
      f, t,       reshaped feature and truth
      in_shape    input shape for model (timesteps x feature)
      num_classes number of classes in truth = output length of nn model
      cweights    weights of each class in truth, per sklearn compute_weights()
      msg         string with report with info on data and operations done
    """
    from sonusai import SonusAIError
    from sonusai.metrics import calculate_class_weights_from_truth

    f = feature
    t = truth

    frames, strides, bands = f.shape
    truth_frames, num_classes = t.shape
    if frames != truth_frames:  # Double-check correctness of inputs
        raise SonusAIError('Frames in feature and truth do not match')

    msg = f'Training/truth shape: {frames}x{strides}x{bands}, nclass/outlen = {num_classes}\n'
    msg += f'Reshape request: timesteps {timesteps}, batchsize {batch_size}, flatten={flatten}, add1ch={add1ch}\n'

    # Compute class weights by hand as sklearn does not handle non-existent classes
    cweights = calculate_class_weights_from_truth(t)

    # calc new input shape only and return
    if batch_size == -1:
        if flatten:
            in_shape = [strides * bands]
        else:
            in_shape = [strides, bands]

        if timesteps > 0:
            in_shape = np.concatenate(([timesteps], in_shape[0:]), axis=0)

        if add1ch:
            in_shape = np.concatenate((in_shape[0:], [1]), axis=0)

        return f, t, in_shape, num_classes, cweights, msg  # quick

    if flatten:
        msg += f'Flattening {strides}x{bands} feature to {strides * bands}\n'
        f = np.reshape(f, (frames, strides * bands))

    # Reshape for Keras/TF recurrent models that require timesteps/sequence length dimension
    if timesteps > 0:
        sequences = frames // timesteps

        # Remove frames if remainder, not fitting into a multiple of new number of sequences
        frem = frames % timesteps
        brem = (frames // timesteps) % batch_size
        bfrem = brem * timesteps
        sequences = sequences - brem
        fr2drop = frem + bfrem
        if fr2drop:
            msg += f'Dropping {fr2drop} frames for new number of sequences to fit in multiple of batch_size\n'
            if f.ndim == 2:
                f = f[0:-fr2drop, ]  # Flattened input
            elif f.ndim == 3:
                f = f[0:-fr2drop, ]  # Un-flattened input

            t = t[0:-fr2drop, ]

        # Do the reshape
        msg += f'Reshape for timesteps = {timesteps}, new number of sequences (batches) = {sequences}\n'
        if f.ndim == 2:  # Flattened input
            # str=str+'Reshaping 2 dim\n'
            f = np.reshape(f, (sequences, timesteps, strides * bands))  # was frames x bands*timesteps
            t = np.reshape(t, (sequences, timesteps, num_classes))  # was frames x num_classes
        elif f.ndim == 3:  # Unflattened input
            # str=str+'Reshaping 3 dim\n'
            f = np.reshape(f, (sequences, timesteps, strides, bands))  # was frames x bands x timesteps
            t = np.reshape(t, (sequences, timesteps, num_classes))  # was frames x num_classes
    else:
        # Drop frames if remainder, not fitting into a multiple of new # sequences (Keras req)
        fr2drop = f.shape[0] % batch_size
        if fr2drop > 0:
            msg += f'Dropping {fr2drop} frames for total to be a multiple of batch_size\n'
            f = f[0:-fr2drop, ]
            t = t[0:-fr2drop, ]

    # Add channel dimension if required for input to model (i.e. for cnn type input)
    if add1ch:
        msg += 'Adding channel dimension to feature\n'
        f = np.expand_dims(f, axis=f.ndim)  # add as last/outermost dim

    in_shape = f.shape
    in_shape = in_shape[1:]  # remove frame dim size

    msg += f'Feature final shape: {f.shape}\n'
    msg += f'Input shape final (includes timesteps): {in_shape}\n'
    msg += f'Truth final shape: {t.shape}\n'

    return f, t, in_shape, num_classes, cweights, msg


def reshape_truth_predict(truth: np.ndarray,
                          predict: np.ndarray,
                          timesteps: int) -> (np.ndarray, np.ndarray, int):
    """Reshape truth and predict data.

    truth and predict can be either frames x num_classes, or frames x timesteps x num_classes
    In binary case, num_classes dim may not exist; detect this and set num_classes == 1
    """
    if truth.ndim == 3 or (truth.ndim == 2 and timesteps > 0):
        if truth.ndim == 2:
            # 2D but has timesteps = truth.shape[1]
            num_classes = 1
        else:
            # frames = truth.shape[0], timesteps = truth.shape[1]
            num_classes = truth.shape[2]
        # reshape to remove timestep dimension
        truth = np.reshape(truth, (truth.shape[0] * truth.shape[1], truth.shape[2]))
        predict = np.reshape(predict, (predict.shape[0] * predict.shape[1], predict.shape[2]))
    else:
        if truth.ndim == 1:
            # no timesteps dimension, = 0
            num_classes = 1
            # convert to 2D (F,1) if only 1 dim
            truth = np.expand_dims(truth, 1)
        else:
            num_classes = truth.shape[1]  # 2D input frames x num_classes

    # Support predict (and truth) when shape is (F,), just convert to (F,1)
    if predict.ndim == 1:
        predict = np.expand_dims(predict, 1)

    return truth, predict, num_classes
